@@ -15,6 +15,8 @@ pub enum StackMachineError {
     NumericOverflow,
     NumberStackUnderflow,
     LoopStackUnderflow,
+    ScratchStackUnderflow,
+    InvalidCellOperation,
     UnhandledTrap,
     RanOutOfGas,
 }
@@ -81,8 +83,9 @@ pub enum Opcode {
     CMPZ,
     CMPNZ,
     LDI(i64),
-    POP,
+    DROP,
     SWAP,
+    SWAP2,
     RET,
     ADD,
     SUB,
@@ -90,6 +93,7 @@ pub enum Opcode {
     DIV,
     NOT,
     DUP,
+    DUP2,
     TRAP,
     NOP,
     PUSHLP,
@@ -99,13 +103,26 @@ pub enum Opcode {
     GETLP2,
     DROPLP,
     CMPLOOP,
+    OVER2,
+    GtR,
+    RGt,
+    RAt,
+    GtR2,
+    RGt2,
+    RAt2,
+    AND,
+    NEWCELLS,
+    MOVETOCELLS,
+    MOVEFROMCELLS,
 }
 
 pub struct StackMachineState {
     pub number_stack: Vec<i64>,
+    pub scratch_stack: Vec<i64>,
     return_stack: Vec<usize>,
     // current index, max_index
     loop_stack: Vec<(i64, i64)>,
+    cells: Vec<i64>,
     pub opcodes: Vec<Opcode>,
     pc: usize,
     gas_used: u64,
@@ -115,8 +132,10 @@ impl Default for StackMachineState {
     fn default() -> Self {
         StackMachineState {
             number_stack: Vec::new(),
+            scratch_stack: Vec::new(),
             return_stack: Vec::new(),
             loop_stack: Vec::new(),
+            cells: Vec::new(),
             opcodes: Vec::new(),
             pc: 0,
             gas_used: 0,
@@ -144,6 +163,48 @@ impl Default for StackMachine {
     }
 }
 
+macro_rules! pop_number_stack {
+    ($variable:ident) => {
+        $variable
+            .st
+            .number_stack
+            .pop()
+            .ok_or(StackMachineError::NumberStackUnderflow)?
+    };
+}
+
+macro_rules! push_number_stack {
+    ($variable:ident,$expr:expr) => {
+        $variable.st.number_stack.push($expr);
+    };
+}
+
+macro_rules! pop_scratch_stack {
+    ($variable:ident) => {
+        $variable
+            .st
+            .scratch_stack
+            .pop()
+            .ok_or(StackMachineError::ScratchStackUnderflow)?
+    };
+}
+
+macro_rules! push_scratch_stack {
+    ($variable:ident,$expr:expr) => {
+        $variable.st.scratch_stack.push($expr);
+    };
+}
+
+macro_rules! last_scratch_stack {
+    ($variable:ident) => {
+        $variable
+            .st
+            .scratch_stack
+            .last()
+            .ok_or(StackMachineError::ScratchStackUnderflow)?
+    };
+}
+
 impl StackMachine {
     /// JR(*) is relative from the JR(*) instruction,
     /// 0 would jump back onto the JR instruction
@@ -166,99 +227,54 @@ impl StackMachine {
             let mut pc_reset = false;
             match self.st.opcodes[self.st.pc] {
                 Opcode::JMP => {
-                    self.st.pc = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .map(|x| x as usize)
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    self.st.pc = usize::try_from(pop_number_stack!(self)).unwrap();
                     pc_reset = true;
                 }
                 Opcode::JR => {
-                    let new_offset = i64::try_from(self.st.pc)?
-                        + self
-                            .st
-                            .number_stack
-                            .pop()
-                            .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let new_offset = i64::try_from(self.st.pc)? + pop_number_stack!(self);
                     self.st.pc = usize::try_from(new_offset).unwrap();
                     pc_reset = true;
                 }
                 Opcode::CALL => {
                     self.st.return_stack.push(self.st.pc + 1);
-                    self.st.pc = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .map(|x| x as usize)
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    self.st.pc = usize::try_from(pop_number_stack!(self))?;
                     pc_reset = true;
                 }
                 Opcode::CMPZ => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let x = pop_number_stack!(self);
                     if x == 0 {
-                        self.st.number_stack.push(0);
-                    } else {
                         self.st.number_stack.push(-1);
+                    } else {
+                        self.st.number_stack.push(0);
                     }
                 }
                 Opcode::CMPNZ => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let x = pop_number_stack!(self);
                     if x == 0 {
-                        self.st.number_stack.push(-1);
-                    } else {
                         self.st.number_stack.push(0);
+                    } else {
+                        self.st.number_stack.push(-1);
                     }
                 }
                 Opcode::JRZ => {
-                    let new_offset = i64::try_from(self.st.pc)?
-                        + self
-                            .st
-                            .number_stack
-                            .pop()
-                            .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let new_offset = i64::try_from(self.st.pc)? + pop_number_stack!(self);
+                    let x = pop_number_stack!(self);
                     if x == 0 {
                         self.st.pc = usize::try_from(new_offset).unwrap();
                         pc_reset = true;
                     }
                 }
                 Opcode::JRNZ => {
-                    let new_offset = i64::try_from(self.st.pc)?
-                        + self
-                            .st
-                            .number_stack
-                            .pop()
-                            .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let new_offset = i64::try_from(self.st.pc)? + pop_number_stack!(self);
+                    let x = pop_number_stack!(self);
                     if x != 0 {
                         self.st.pc = usize::try_from(new_offset).unwrap();
                         pc_reset = true;
                     }
                 }
-                Opcode::LDI(x) => self.st.number_stack.push(x),
-                Opcode::POP => {
-                    let _ = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                Opcode::LDI(x) => push_number_stack!(self, x),
+                Opcode::DROP => {
+                    let _ = pop_number_stack!(self);
                 }
                 Opcode::RET => {
                     match self.st.return_stack.pop() {
@@ -267,98 +283,111 @@ impl StackMachine {
                     };
                     pc_reset = true;
                 }
+                Opcode::GtR => {
+                    let x = pop_number_stack!(self);
+                    push_scratch_stack!(self, x);
+                }
+                Opcode::RGt => {
+                    let x = pop_scratch_stack!(self);
+                    push_number_stack!(self, x);
+                }
+                Opcode::RAt => {
+                    let x = last_scratch_stack!(self);
+                    push_number_stack!(self, *x);
+                }
+                Opcode::GtR2 => {
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_scratch_stack!(self, y);
+                    push_scratch_stack!(self, x);
+                }
+                Opcode::RGt2 => {
+                    let x = pop_scratch_stack!(self);
+                    let y = pop_scratch_stack!(self);
+                    push_number_stack!(self, y);
+                    push_number_stack!(self, x);
+                }
+                Opcode::RAt2 => {
+                    let x = pop_scratch_stack!(self);
+                    let y = pop_scratch_stack!(self);
+                    push_scratch_stack!(self, y);
+                    push_scratch_stack!(self, x);
+                    push_number_stack!(self, y);
+                    push_number_stack!(self, x);
+                }
                 Opcode::ADD => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let y = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(x + y);
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, x + y);
                 }
                 Opcode::SUB => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let y = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(x - y);
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, x - y);
                 }
                 Opcode::MUL => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let y = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(x * y);
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, x * y);
                 }
                 Opcode::DIV => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let y = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(x / y);
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, y / x);
                 }
                 Opcode::NOT => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(match x {
-                        0 => 1,
-                        _ => 0,
-                    });
+                    let x = pop_number_stack!(self);
+                    push_number_stack!(
+                        self,
+                        match x {
+                            0 => 1,
+                            _ => 0,
+                        }
+                    );
                 }
                 Opcode::DUP => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(x);
-                    self.st.number_stack.push(x);
+                    let x = pop_number_stack!(self);
+                    push_number_stack!(self, x);
+                    push_number_stack!(self, x);
+                }
+                Opcode::DUP2 => {
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, y);
+                    push_number_stack!(self, x);
+                    push_number_stack!(self, y);
+                    push_number_stack!(self, x);
+                }
+                Opcode::OVER2 => {
+                    let x4 = pop_number_stack!(self);
+                    let x3 = pop_number_stack!(self);
+                    let x2 = pop_number_stack!(self);
+                    let x1 = pop_number_stack!(self);
+                    push_number_stack!(self, x1);
+                    push_number_stack!(self, x2);
+                    push_number_stack!(self, x3);
+                    push_number_stack!(self, x4);
+                    push_number_stack!(self, x1);
+                    push_number_stack!(self, x2);
                 }
                 Opcode::SWAP => {
-                    let x = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let y = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    self.st.number_stack.push(x);
-                    self.st.number_stack.push(y);
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, x);
+                    push_number_stack!(self, y);
+                }
+                Opcode::SWAP2 => {
+                    let x4 = pop_number_stack!(self);
+                    let x3 = pop_number_stack!(self);
+                    let x2 = pop_number_stack!(self);
+                    let x1 = pop_number_stack!(self);
+                    push_number_stack!(self, x3);
+                    push_number_stack!(self, x4);
+                    push_number_stack!(self, x1);
+                    push_number_stack!(self, x2);
                 }
                 Opcode::TRAP => {
-                    let trap_id = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let trap_id = pop_number_stack!(self);
                     for h in self.trap_handlers.iter_mut() {
                         if let TrapHandled::Handled = h.handle_trap(trap_id, &mut self.st)? {
                             return Ok(());
@@ -368,16 +397,8 @@ impl StackMachine {
                 }
                 Opcode::NOP => {}
                 Opcode::PUSHLP => {
-                    let current_index = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
-                    let max_index = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let current_index = pop_number_stack!(self);
+                    let max_index = pop_number_stack!(self);
                     self.st.loop_stack.push((current_index, max_index));
                 }
                 Opcode::INCLP => match self.st.loop_stack.last_mut() {
@@ -389,11 +410,7 @@ impl StackMachine {
                     }
                 },
                 Opcode::ADDLP => {
-                    let increment = self
-                        .st
-                        .number_stack
-                        .pop()
-                        .ok_or(StackMachineError::NumberStackUnderflow)?;
+                    let increment = pop_number_stack!(self);
 
                     match self.st.loop_stack.last_mut() {
                         Some((current_index, _max_index)) => {
@@ -442,6 +459,43 @@ impl StackMachine {
                         self.st.number_stack.push(0);
                     }
                 }
+                Opcode::AND => {
+                    let x = pop_number_stack!(self);
+                    let y = pop_number_stack!(self);
+                    push_number_stack!(self, x & y);
+                }
+                Opcode::NEWCELLS => {
+                    let num_cells = usize::try_from(pop_number_stack!(self))
+                        .map_err(|_| StackMachineError::InvalidCellOperation)?;
+                    let newaddress = self.st.cells.len();
+                    self.st
+                        .cells
+                        .resize_with(newaddress + num_cells, Default::default);
+                }
+                Opcode::MOVETOCELLS => {
+                    let num_cells = usize::try_from(pop_number_stack!(self))
+                        .map_err(|_| StackMachineError::InvalidCellOperation)?;
+                    let address = usize::try_from(pop_number_stack!(self))
+                        .map_err(|_| StackMachineError::InvalidCellOperation)?;
+                    if num_cells < 1 || self.st.cells.len() < address + num_cells {
+                        return Err(StackMachineError::InvalidCellOperation);
+                    }
+                    for i in address..address + num_cells {
+                        self.st.cells[i] = pop_number_stack!(self);
+                    }
+                }
+                Opcode::MOVEFROMCELLS => {
+                    let num_cells = usize::try_from(pop_number_stack!(self))
+                        .map_err(|_| StackMachineError::InvalidCellOperation)?;
+                    let address = usize::try_from(pop_number_stack!(self))
+                        .map_err(|_| StackMachineError::InvalidCellOperation)?;
+                    if num_cells < 1 || self.st.cells.len() < address + num_cells {
+                        return Err(StackMachineError::InvalidCellOperation);
+                    }
+                    for i in (address..address + num_cells).rev() {
+                        push_number_stack!(self, self.st.cells[i]);
+                    }
+                }
             };
             if !pc_reset {
                 self.st.pc += 1;
