@@ -191,7 +191,7 @@ fn test_execute_cmpz_1() {
     // Execute the instructions
     sm.execute(0, GasLimit::Limited(100)).unwrap();
 
-    assert_eq!(sm.st.number_stack, vec![123_i64, 321, 0]);
+    assert_eq!(sm.st.number_stack, vec![123_i64, 321, -1]);
 }
 
 #[test]
@@ -208,7 +208,7 @@ fn test_execute_cmpz_2() {
     // Execute the instructions
     sm.execute(0, GasLimit::Limited(100)).unwrap();
 
-    assert_eq!(sm.st.number_stack, vec![123_i64, 321, -1]);
+    assert_eq!(sm.st.number_stack, vec![123_i64, 321, 0]);
 }
 
 #[test]
@@ -225,7 +225,7 @@ fn test_execute_cmpnz_1() {
     // Execute the instructions
     sm.execute(0, GasLimit::Limited(100)).unwrap();
 
-    assert_eq!(sm.st.number_stack, vec![123_i64, 321, -1]);
+    assert_eq!(sm.st.number_stack, vec![123_i64, 321, 0]);
 }
 
 #[test]
@@ -242,7 +242,7 @@ fn test_execute_cmpnz_2() {
     // Execute the instructions
     sm.execute(0, GasLimit::Limited(100)).unwrap();
 
-    assert_eq!(sm.st.number_stack, vec![123_i64, 321, 0]);
+    assert_eq!(sm.st.number_stack, vec![123_i64, 321, -1]);
 }
 
 #[test]
@@ -290,6 +290,113 @@ fn test_execute_call() {
     );
 }
 
+#[test]
+fn test_execute_gt_r() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0, 1, 2]);
+    // Put the opcodes into the *memory*
+    sm.st.opcodes.extend_from_slice(&[Opcode::GtR, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1]);
+    assert_eq!(sm.st.scratch_stack, vec![2_i64]);
+}
+
+#[test]
+fn test_execute_r_gt() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0, 1, 2]);
+    sm.st.scratch_stack.extend_from_slice(&[3, 4, 5]);
+    // Put the opcodes into the *memory*
+    sm.st.opcodes.extend_from_slice(&[Opcode::RGt, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1, 2, 5]);
+    assert_eq!(sm.st.scratch_stack, vec![3_i64, 4]);
+}
+
+#[test]
+fn test_execute_r_at() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0, 1, 2]);
+    sm.st.scratch_stack.extend_from_slice(&[3, 4, 5]);
+    // Put the opcodes into the *memory*
+    sm.st.opcodes.extend_from_slice(&[Opcode::RAt, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1, 2, 5]);
+    assert_eq!(sm.st.scratch_stack, vec![3_i64, 4, 5]);
+}
+
+#[test]
+fn test_execute_gt_r_2() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0, 1, 2]);
+    sm.st.scratch_stack.extend_from_slice(&[3, 4, 5]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::GtR2, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64]);
+    assert_eq!(sm.st.scratch_stack, vec![3_i64, 4, 5, 1, 2]);
+}
+
+#[test]
+fn test_execute_r_gt_2() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0, 1, 2]);
+    sm.st.scratch_stack.extend_from_slice(&[3, 4, 5]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::RGt2, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1, 2, 4, 5]);
+    assert_eq!(sm.st.scratch_stack, vec![3_i64]);
+}
+
+#[test]
+fn test_execute_r_at_2() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0, 1, 2]);
+    sm.st.scratch_stack.extend_from_slice(&[3, 4, 5]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::RAt2, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1, 2, 4, 5]);
+    assert_eq!(sm.st.scratch_stack, vec![3_i64, 4, 5]);
+}
+
 #[test]
 fn test_execute_ldi() {
     let mut sm = StackMachine::default();
@@ -317,7 +424,7 @@ fn test_execute_pop() {
     sm.st.opcodes.extend_from_slice(&[
         Opcode::LDI(0),
         Opcode::LDI(1),
-        Opcode::POP,
+        Opcode::DROP,
         Opcode::LDI(2),
         Opcode::RET,
     ]);
@@ -339,11 +446,11 @@ fn test_execute_pop_error() {
     sm.st.opcodes.extend_from_slice(&[
         Opcode::LDI(0),
         Opcode::LDI(1),
-        Opcode::POP,
-        Opcode::POP,
-        Opcode::POP,
-        Opcode::POP,
-        Opcode::POP,
+        Opcode::DROP,
+        Opcode::DROP,
+        Opcode::DROP,
+        Opcode::DROP,
+        Opcode::DROP,
         Opcode::LDI(2),
         Opcode::RET,
     ]);
@@ -423,14 +530,14 @@ fn test_execute_div() {
     let mut sm = StackMachine::default();
 
     // Populate the number stack
-    sm.st.number_stack.extend_from_slice(&[321, 39483]);
+    sm.st.number_stack.extend_from_slice(&[10, 5]);
     // Put the opcodes into the *memory*
     sm.st.opcodes.extend_from_slice(&[Opcode::DIV, Opcode::RET]);
 
     // Execute the instructions
     sm.execute(0, GasLimit::Limited(100)).unwrap();
 
-    assert_eq!(sm.st.number_stack, vec![123]);
+    assert_eq!(sm.st.number_stack, vec![2]);
 }
 
 #[test]
@@ -831,3 +938,272 @@ fn test_execute_cmpgelp_lt() {
     assert_eq!(sm.st.number_stack, vec![321, 39583, 0]);
     assert_eq!(sm.st.loop_stack, vec![(3210, 0), (39482, 39483)]);
 }
+
+#[test]
+fn test_execute_and() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0b10101110i64, 0b01010111i64]);
+    // Put the opcodes into the *memory*
+    sm.st.opcodes.extend_from_slice(&[Opcode::AND, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0b00000110i64]);
+}
+
+#[test]
+fn test_execute_newcells_1() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0_i64, 2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::NEWCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0]);
+    assert_eq!(sm.st.cells, vec![0, 0]);
+}
+
+#[test]
+fn test_execute_newcells_2() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    sm.st.number_stack.extend_from_slice(&[0_i64, -2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::NEWCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
+
+#[test]
+fn test_execute_movetocells_1() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // 2 is the number of values to move to cells
+    // 0 is the location to start moving values to
+    // 3 2 1 are the values to use when moving to cells
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, 0, 2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVETOCELLS, Opcode::RET]);
+
+    // Setup the cells we will be storing to
+    sm.st.cells.extend_from_slice(&[0, 0]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1]);
+    assert_eq!(sm.st.cells, vec![3, 2]);
+}
+
+#[test]
+fn test_execute_movetocells_2() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // -2 Use an invalid number for the number of cells to cause a fault
+    // 0 is the start location to start
+    // 0 is the location to start moving values to
+    // 3 2 1 are the values to use when moving to cells
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, 0, -2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVETOCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
+
+#[test]
+fn test_execute_movetocells_3() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // 2 is the number of values to move to cells
+    // -5 is an invalid start location to cause a fault
+    // 0 is the location to start moving values to
+    // 3 2 1 are the values to use when moving to cells
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, -5, 2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVETOCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
+
+#[test]
+fn test_execute_movetocells_4() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // 3 is the number of values to move to cells, it should cause a fault
+    // 0 is the location to start moving values to
+    // 3 2 1 are the values to use when moving to cells
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, 0, 3]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVETOCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
+
+#[test]
+fn test_execute_movefromcells_1() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // 2 is the number of values to move to cells
+    // 0 is the location to start moving values to
+    // 3 2 1 are the values to use when moving to cells
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, 0, 2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVEFROMCELLS, Opcode::RET]);
+
+    // Setup the cells we will be storing to
+    sm.st.cells.extend_from_slice(&[5, 4]);
+
+    // Execute the instructions
+    sm.execute(0, GasLimit::Limited(100)).unwrap();
+
+    assert_eq!(sm.st.number_stack, vec![0_i64, 1, 2, 3, 4, 5]);
+    assert_eq!(sm.st.cells, vec![5, 4]);
+}
+
+#[test]
+fn test_execute_movefromcells_2() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // -2 Use an invalid number for the number of cells to cause a fault
+    // 0 is the start location to start
+    // 0 is the location to start moving values from
+    // 3 2 1 are the values left on the stack
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, 0, -2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVEFROMCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
+
+#[test]
+fn test_execute_movefromcells_3() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // 2 is the number of values to move from cells
+    // -5 is an invalid start location to cause a fault
+    // 0 is the location to start moving values from
+    // 3 2 1 are the values left on the stack
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, -5, 2]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVEFROMCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
+
+#[test]
+fn test_execute_movefromcells_4() {
+    let mut sm = StackMachine::default();
+
+    // Populate the number stack
+    // 3 is the number of values to move from cells, it should cause a fault
+    // 0 is a start location
+    // 0 is the location to start moving values from
+    // 3 2 1 are the values left on the stack
+    sm.st
+        .number_stack
+        .extend_from_slice(&[0_i64, 1, 2, 3, 0, 3]);
+    // Put the opcodes into the *memory*
+    sm.st
+        .opcodes
+        .extend_from_slice(&[Opcode::MOVEFROMCELLS, Opcode::RET]);
+
+    // Execute the instructions
+    assert_eq!(
+        match sm.execute(0, GasLimit::Limited(100)) {
+            Err(StackMachineError::InvalidCellOperation) => 1,
+            _ => 0,
+        },
+        1
+    );
+}
